"""
Utility functions writing and reading BUFR files from AWS data

see documentation here:
https://confluence.ecmwf.int/display/ECC/Documentation

BUFR element table for WMO master table version 32
https://confluence.ecmwf.int/display/ECC/WMO%3D32+element+table

"""
import datetime
import logging
import math
from os import PathLike
from pathlib import Path
from typing import BinaryIO, Optional

import attrs
import numpy as np
import pandas as pd
from eccodes import (
    codes_set,
    codes_write,
    codes_release,
    codes_bufr_new_from_samples,
    CodesInternalError,
    codes_is_defined,
    codes_bufr_new_from_file,
    codes_get,
)

logger = logging.getLogger(__name__)

__all__ = [
    "BUFRVariables",
    "write_bufr_message",
    "read_bufr_message",
    "read_bufr_file",
]


def round_converter(decimals: int):
    def round(value: float):
        return np.round(value, decimals=decimals)

    return round


# Enforce precision
# Note the sensor accuracies listed here:
# https://essd.copernicus.org/articles/13/3819/2021/#section8
# In addition to sensor accuracy, WMO requires pressure and heights
# to be reported at 0.1 precision.
@attrs.define(eq=False)
class BUFRVariables:
    """
    Helper class for storing variables used for BUFR IO.

    The field names reflect the key names in the BUFR template except:

    * wmo_id: Stored as either as shipOrMobileLandStationIdentifier or stationNumber depending on the station type
    * station_type: Determine the BUFR template
    * timestamp: Stored separately as year, month, day, hour and minutes
    * heightOfSensorAboveLocalGroundOrDeckOfMarinePlatformTempRH: Corresponds to "#1#heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform" which is height if thermometer and hygrometer relative to ground or deck of marine platform.
    * heightOfSensorAboveLocalGroundOrDeckOfMarinePlatformWSPD: Corresponds to "#7#heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform" which is height if anemometer relative to ground or deck of marine platform.

    """

    # Station type: "mobile" or "land"
    # ===============================
    # Fixed land station schema: https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_D/307080
    # Mobile station schema: https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_D/307090

    station_type: str

    # WMO station identifier
    # Land stations: https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_D/301090
    # Mobile stations: https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_D/301092
    # ======================================================================================================
    wmo_id: str
    timestamp: datetime.datetime

    # https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_B/005001
    # Scale: 5, unit: degrees
    # TODO: Test if eccodes does the rounding as well. The rounding is was 6 which is larger that the scale.
    latitude: float = attrs.field(converter=round_converter(5))
    # https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_B/006001
    # Scale: 5, unit: degrees
    longitude: float = attrs.field(converter=round_converter(5))

    # https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_B/007030
    # Scale: 1, unit: m
    heightOfStationGroundAboveMeanSeaLevel: float = attrs.field(
        converter=round_converter(1)
    )
    # https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_B/007031
    # Scale: 1, unit: m
    heightOfBarometerAboveMeanSeaLevel: float = attrs.field(
        converter=round_converter(1),
    )

    # Pressure information
    # ====================
    # Definition table: https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_D/302031
    # https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_B/010004
    # Scale: -1, unit: Pa
    nonCoordinatePressure: float = attrs.field(converter=round_converter(-1))
    # There are two other pressure variables in the template: 007004 - pressure and 010062 24-hour pressure change

    # Basic synoptic "instantaneous" data
    # ===================================
    # Definition table: https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_D/302035
    # This section only include the temperature and humidity data (302032).
    # Precipitation and cloud data are currently ignored.
    # https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_B/007032
    # Scale: 2, unit: m
    # This is the first appearance of this variable id.
    heightOfSensorAboveLocalGroundOrDeckOfMarinePlatformTempRH: float = attrs.field(
        converter=round_converter(2),
    )
    # https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_B/012101
    # Scale: 2, unit: K
    # NOTE: The expected scale is 2, but our instantanous data is rounded to 1 decimal.
    airTemperature: float = attrs.field(converter=round_converter(1))
    # There is also a Dewpoint temperature in this template: 012103 which is currently unused.
    # https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_B/012103
    # Scale: 0, unit: %
    relativeHumidity: float = attrs.field(converter=round_converter(0))

    # Basic synoptic "period" data
    # ============================
    # Definition table: https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_D/302043
    # Wind data: https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_D/302042
    # Wind direction: https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_B/011001
    # Scale: 0, unit: degrees
    windDirection: float = attrs.field(converter=round_converter(0))
    # Wind speed: https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_B/011002
    # Scale: 1, unit: m/s
    windSpeed: float = attrs.field(converter=round_converter(1))
    # https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_B/007032
    # Scale: 2, unit: m
    # This is the 7th appearance of this variable id.
    heightOfSensorAboveLocalGroundOrDeckOfMarinePlatformWSPD: float = attrs.field(
        converter=round_converter(2)
    )

    def as_series(self) -> pd.Series:
        return pd.Series(attrs.asdict(self))

    def __eq__(self, other: "BUFRVariables"):
        """Use pandas series equals to allow nan values in comparison."""
        return self.as_series().equals(other.as_series())


STATION_CONFIGURATIONS = {
    "mobile": {
        # 'blockNumber': 4, #4 is Greenland, 6 is Denmark; not valid with synopMobil template
        "regionNumber": 6,  # 6 is Europe, 7 is MISSING VALUE; not valid with synopLand template
        "centre": 94,  # 94 is Copenhagen
        # 'agencyInChargeOfOperatingObservingPlatform': , #nothing for DMI or GEUS in code table
        # 'wmoRegionSubArea': 1,
        # 'stationOrSiteName': , #not valid with synopMobil template
        # 'shortStationName': , #not valid with synopMobil template
        # 'longStationName': , #not valid with synopMobil template
        # 'directionOfMotionOfMovingObservingPlatform': ,
        # 'movingObservingPlatformSpeed': ,
        "stationType": 0,  # automatic station
        "instrumentationForWindMeasurement": 8,  # certified instruments
        "stationElevationQualityMarkForMobileStations": 1,  # Excellent - within 3m; not valid with synopLand template
    },
    "land": {
        "blockNumber": 4,  # 4 is Greenland, 6 is Denmark; not valid with synopMobil template
        # 'regionNumber': 6, #6 is Europe, 7 is MISSING VALUE; not valid with synopLand template
        "centre": 94,  # 94 is Copenhagen
        # 'agencyInChargeOfOperatingObservingPlatform': , #nothing for DMI or GEUS in code table
        # 'wmoRegionSubArea': 1,
        # 'stationOrSiteName': , #not valid with synopMobil template
        # 'shortStationName': , #not valid with synopMobil template
        # 'longStationName': , #not valid with synopMobil template
        "stationType": 0,  # automatic station
        "instrumentationForWindMeasurement": 8,  # certified instruments
        # 'stationElevationQualityMarkForMobileStations': 1, #Excellent - within 3m; not valid with synopLand template
    },
}

BUFR_TEMPLATES = {
    "mobile": {
        # Template definition: https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_D/307090
        "unexpandedDescriptors": (307090),  # message template, "synopMobil"
        "edition": 4,  # latest edition
        "masterTableNumber": 0,
        "masterTablesVersionNumber": 32,  # DMI recommends any table version between 28-32
        "localTablesVersionNumber": 0,
        "bufrHeaderCentre": 94,  # originating centre 98=ECMWF, 94=DMI
        # 'bufrHeaderSubCentre': 0,
        "updateSequenceNumber": 0,  # 0 is original message, incremented by 1 for updates
        "dataCategory": 0,  # surface data - land
        "internationalDataSubCategory": 3,  # hourly synoptic observations from mobile-land stations (SYNOP MOBIL)
        # 'dataSubCategory': 0,
        "observedData": 1,
        "compressedData": 0,
    },
    "land": {
        # Template definition: https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_D/307080
        "unexpandedDescriptors": (307080),  # message template, "synopLand"
        "edition": 4,  # latest edition
        "masterTableNumber": 0,
        "masterTablesVersionNumber": 32,  # DMI recommends any table version between 28-32
        "localTablesVersionNumber": 0,
        "bufrHeaderCentre": 94,  # originating centre 98=ECMWF, 94=DMI
        # 'bufrHeaderSubCentre': 0,
        "updateSequenceNumber": 0,  # 0 is original message, incremented by 1 for updates
        "dataCategory": 0,  # surface data - land
        "internationalDataSubCategory": 0,  # Hourly synoptic observations from fixed-land stations (SYNOP)
        # 'dataSubCategory': 0,
        "observedData": 1,
        "compressedData": 0,
    },
}


def write_bufr_message(
    variables: BUFRVariables,
    file: BinaryIO,
):
    """Construct and export .bufr message to file from pandas Series.

    Parameters
    ----------
    variables : pandas.Series
        Pandas series of single most recent obset for a station
    file
        Binary writable file object
    """

    # Create new bufr message to write to
    ibufr = codes_bufr_new_from_samples("BUFR4")

    try:
        # we must pass all the following functions without error.
        # If handled (or unhandled) errors occur, we re-raise and
        # the exceptions below will set remove_file to True.
        set_template(ibufr, variables.timestamp, variables.station_type)
        set_station(ibufr, variables.station_type, variables.wmo_id)
        set_AWS_variables(ibufr, variables)

        # Encode keys in data section
        codes_set(ibufr, "pack", 1)

        # Write bufr message to bufr file
        codes_write(ibufr, file)

    except CodesInternalError as ec:
        logger.exception(f"CodesInternalError in getBUFR", exc_info=ec)
        raise ec
    except Exception as e:
        logger.exception(f"ERROR in getBUFR", exc_info=e)
        raise e
    finally:
        codes_release(ibufr)


def set_template(ibufr, timestamp, station_type: str):
    """Set BUFR message template.

    Parameters
    ----------
    ibufr : bufr.msg
        Bufr message object
    timestamp : datetime.Datetime
        Timestamp of observation
    config_key : str
        Defines which config dict to use in wmo_config.ibufr_settings, 'mobile' or 'land'
    """
    template = BUFR_TEMPLATES[station_type]

    for k, v in template.items():
        if codes_is_defined(ibufr, k) == 1:
            codes_set(ibufr, k, v)
        else:
            logger.warning("-----> setTemplate Key not defined: {}".format(k))
            continue

    codes_set(ibufr, "typicalYear", timestamp.year)
    codes_set(ibufr, "typicalMonth", timestamp.month)
    codes_set(ibufr, "typicalDay", timestamp.day)
    codes_set(ibufr, "typicalHour", timestamp.hour)
    codes_set(ibufr, "typicalMinute", timestamp.minute)
    # codes_set(ibufr, 'typicalSecond', timestamp.second)


def set_station(ibufr, station_type: str, wmo_id: str):
    """Set station-specific info to bufr message.

    Parameters
    ----------
    ibufr : bufr.msg
        Bufr message object
    config_key : str
        Defines which config dict to use in wmo_config.ibufr_settings, 'mobile' or 'land'
    """
    if station_type == "mobile":
        station_config = dict(shipOrMobileLandStationIdentifier=wmo_id)
    elif station_type == "land":
        # StationNumber for land stations are integeres
        wmo_id_int = int(wmo_id)
        if wmo_id_int >= 1024:
            raise ValueError(
                f"Invalid WMO ID {wmo_id}. Land station number must be less than 1024."
                "See https://vocabulary-manager.eumetsat.int/vocabularies/BUFR/WMO/32/TABLE_B/001002"
            )
        station_config = dict(stationNumber=wmo_id_int)
    else:
        raise Exception(f"Unsupported station station type {station_type}")
    station_config.update(STATION_CONFIGURATIONS[station_type])

    for key, value in station_config.items():
        codes_set(ibufr, key, value)


def set_AWS_variables(
    ibufr,
    variables: BUFRVariables,
):
    """Set AWS measurements to bufr message.

    Parameters
    ----------
    ibufr s: bufr.msg
        Bufr message object
    variables
        Dict with AWS variable data
    timestamp : datetime.datetime
        timestamp for this row
    """
    # Set timestamp fields
    timestamp = variables.timestamp
    set_bufr_value(ibufr, "year", timestamp.year)
    set_bufr_value(ibufr, "month", timestamp.month)
    set_bufr_value(ibufr, "day", timestamp.day)
    set_bufr_value(ibufr, "hour", timestamp.hour)
    set_bufr_value(ibufr, "minute", timestamp.minute)

    set_bufr_value(ibufr, "relativeHumidity", variables.relativeHumidity)
    set_bufr_value(ibufr, "airTemperature", variables.airTemperature)
    set_bufr_value(ibufr, "nonCoordinatePressure", variables.nonCoordinatePressure)
    set_bufr_value(ibufr, "windDirection", variables.windDirection)
    set_bufr_value(ibufr, "windSpeed", variables.windSpeed)

    set_bufr_value(ibufr, "latitude", variables.latitude)

    # Set position metadata
    set_bufr_value(ibufr, "latitude", variables.latitude)
    set_bufr_value(ibufr, "longitude", variables.longitude)
    set_bufr_value(
        ibufr,
        "heightOfStationGroundAboveMeanSeaLevel",
        variables.heightOfStationGroundAboveMeanSeaLevel,
    )  # also height and heightOfStation?

    # The ## in the codes_set() indicate the position in the BUFR for the parameter.
    # e.g. #10#timePeriod will assign to the 10th occurence of "timePeriod", which corresponds
    # to the wind speed section. Note that both the "synopMobil" and "synopLand" templates
    # appear to have the same positions for all parameters that are set here.
    # View the output BUFR to see section keys with 'bufr_dump filename.bufr'.
    if math.isnan(variables.windSpeed) is False:
        # Set time significance (2=temporally averaged)
        codes_set(ibufr, "#1#timeSignificance", 2)
        # Set monitoring time period (-10=10 minutes)
        codes_set(ibufr, "#10#timePeriod", -10)

    # Set measurement heights
    set_bufr_value(
        ibufr,
        "#1#heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform",
        variables.heightOfSensorAboveLocalGroundOrDeckOfMarinePlatformTempRH,
    )
    set_bufr_value(
        ibufr,
        "#7#heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform",
        variables.heightOfSensorAboveLocalGroundOrDeckOfMarinePlatformWSPD,
    )
    set_bufr_value(
        ibufr,
        "heightOfBarometerAboveMeanSeaLevel",
        variables.heightOfBarometerAboveMeanSeaLevel,
    )  # For pressure


def set_bufr_value(ibufr, b_name, value):
    """Set variable in BUFR message
    Called in setAWSvariables() to make sure we aren't passing NaNs

    Parameters
    ----------
    ibufr : bufr.msg
        Active BUFR message
    b_name : str
        BUFR message variable name
    value : int/float
        Value to be assigned to variable
    """
    if math.isnan(value) is False:
        try:
            codes_set(ibufr, b_name, value)
        except CodesInternalError:
            logger.exception(f"CodesInternalError for {b_name} == {value}")
            raise  # throw error back to getBUFR where it is handled
    else:
        logger.info(f"Variable {b_name} is {value}. Skipping")


def get_bufr_value(msgid: int, key: str) -> float:
    """
    Read and convert numeric BUFR values and interpret nan based on value.

    Nan values are skipped in set_bufr_value. This means that they have a default value given by the template.

    * int:  2147483647 == 2**31 -1
    * float: -1e100

    Note: windDirection and relativeHumidity are serialized as integer in the BUFR message.
    """
    value = codes_get(msgid, key)

    if isinstance(value, int):
        if value > 2**30:
            return np.nan
        return value
    elif isinstance(value, float):
        if value == -1e100:
            return np.nan
        return value
    else:
        raise ValueError(f"Unsupported BUFR value type {type(value)} for key {key}")


def read_bufr_message(fp: BinaryIO, backwards_compatible: bool = False) -> Optional[BUFRVariables]:
    """
    Read and parse BUFR message from binary IO stream.

    Extract AWS variables similar to the input to bufr_utilities.write_bufr_message.
    Note: stid is not written to the BUFR file hence it will be set to None in the output.

    Parameters
    ----------
    fp
        Readable binary io stream
    backwards_compatible
        Use legacy pressure if nonCoordinatePressure is nan

    Returns
    -------
    BUFRVariables
        AWS variables or None if there are no messages in stream
    """
    ibufr = codes_bufr_new_from_file(fp)
    if ibufr is None:
        return None
    codes_set(ibufr, "unpack", 1)

    year = codes_get(
        ibufr,
        "year",
    )
    month = codes_get(
        ibufr,
        "month",
    )
    day = codes_get(
        ibufr,
        "day",
    )
    hour = codes_get(
        ibufr,
        "hour",
    )
    minute = codes_get(
        ibufr,
        "minute",
    )
    timestamp = datetime.datetime(
        year=year, month=month, day=day, hour=hour, minute=minute
    )

    # Determine template
    unexpanded_descriptors = codes_get(ibufr, "unexpandedDescriptors")
    if unexpanded_descriptors == 307090:
        # "synopMobil"
        station_type = "mobile"
        wmo_id = codes_get(ibufr, "shipOrMobileLandStationIdentifier")
    elif unexpanded_descriptors == 307080:
        # "synopLand"
        station_type = "land"
        # Note: stationNumber is an integer
        station_number = codes_get(ibufr, "stationNumber")
        wmo_id = str(station_number)
    else:
        raise ValueError(
            f"Unknown BUFR template unexpandedDescriptors: {unexpanded_descriptors}"
        )

    nonCoordinatePressure = get_bufr_value(ibufr, "nonCoordinatePressure")
    if math.isnan(nonCoordinatePressure) and backwards_compatible:
        nonCoordinatePressure = get_bufr_value(ibufr, "pressure")
        if not math.isnan(nonCoordinatePressure):
            logger.warning(
                f"nonCoordinatePressure is nan, using legacy pressure instead"
            )

    variables = BUFRVariables(
        timestamp=timestamp,
        relativeHumidity=get_bufr_value(ibufr, "relativeHumidity"),
        airTemperature=get_bufr_value(ibufr, "airTemperature"),
        nonCoordinatePressure=nonCoordinatePressure,
        windDirection=get_bufr_value(ibufr, "windDirection"),
        windSpeed=get_bufr_value(ibufr, "windSpeed"),
        latitude=get_bufr_value(ibufr, "latitude"),
        longitude=get_bufr_value(ibufr, "longitude"),
        heightOfStationGroundAboveMeanSeaLevel=get_bufr_value(
            ibufr, "heightOfStationGroundAboveMeanSeaLevel"
        ),
        heightOfSensorAboveLocalGroundOrDeckOfMarinePlatformTempRH=get_bufr_value(
            ibufr, "#1#heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform"
        ),
        heightOfSensorAboveLocalGroundOrDeckOfMarinePlatformWSPD=get_bufr_value(
            ibufr, "#7#heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform"
        ),
        heightOfBarometerAboveMeanSeaLevel=get_bufr_value(
            ibufr, "heightOfBarometerAboveMeanSeaLevel"
        ),
        wmo_id=wmo_id,
        station_type=station_type,
    )
    codes_release(ibufr)

    return variables


def read_bufr_file(path: PathLike) -> pd.DataFrame:
    """
    Read aws data from all messages in a bufr file.

    Parameters
    ----------
    path : PathLike
        Path to bufr file

    Returns
    -------
    pd.DataFrame

    """
    path = Path(path)
    lines = []
    with path.open("rb") as fp:
        while True:
            message_vars = read_bufr_message(fp)
            if message_vars is None:
                break
            lines.append(message_vars.as_series())
    data_frame = pd.DataFrame(lines).set_index("wmo_id")
    return data_frame



#!/usr/bin/env python
from argparse import ArgumentParser
import os, unittest
from pypromice.get.get import aws_data


def parse_arguments_data():
	parser = ArgumentParser(description="PROMICE and GC-Net dataset fetcher")       
	parser.add_argument('-n', '--awsname', default=None, type=str, required=True, 
		        help='AWS name')
	parser.add_argument('-f', '--format', default='csv', type=str, required=False, 
		        help='File format to save data as')                      
	parser.add_argument('-o', '--outpath', default=os.getcwd(), type=str, required=False, 
		        help='Directory where file will be written to')          	        
	args = parser.parse_args()
	return args


def get_promice_data():
    '''Command line driver for fetching PROMICE and GC-Net datasets'''

    args = parse_arguments_data()
	
    # Construct AWS dataset name
#    n = aws_names()
#    assert(args.awsname in n) 

    # Check file format type
    f = args.format.lower()
    assert(args.format in ['csv', 'nc', '.csv', '.nc'])
	
    # Construct output file path
    assert(os.path.exists(args.outpath))
	
    # Remove pre-existing files of same name
    if os.path.isfile(f):
        os.remove(f)
	    	
    # Fetch data 
    print(f'Fetching {args.awsname.lower()}...')
    data = aws_data(args.awsname.lower())
	
    # Save to file
    if f in 'csv':
        outfile = os.path.join(args.outpath, args.awsname.lower()) 
        if outfile is not None:
            data.to_csv(outfile)
    elif f in 'nc': 
        data.to_netcdf(outfile, mode='w', format='NETCDF4', compute=True)
        if outfile is not None:
            outfile = os.path.join(args.outpath, args.awsname.lower().split('.csv')[0]+'.nc')
        	
    print(f'File saved to {outfile}')

if __name__ == "__main__":  
    get_promice_data()



#!/usr/bin/env python
import json
import logging, os, sys, toml
from argparse import ArgumentParser

from pypromice.utilities.git import get_commit_hash_and_check_dirty

import pypromice.resources
from pypromice.process.write import prepare_and_write
import numpy as np
import pandas as pd
import xarray as xr

logging.basicConfig(
    format="%(asctime)s; %(levelname)s; %(name)s; %(message)s",
    level=logging.INFO,
    stream=sys.stdout,
)
logger = logging.getLogger(__name__)


def parse_arguments_joinl3(debug_args=None):
    parser = ArgumentParser(
        description="AWS L3 script for the processing L3 data from L2 and merging the L3 data with its historical site. An hourly, daily and monthly L3 data product is outputted to the defined output path"
    )
    parser.add_argument(
        "-c",
        "--config_folder",
        type=str,
        required=True,
        help="Path to folder with sites configuration (TOML) files",
    )
    parser.add_argument(
        "-s",
        "--site",
        default=None,
        type=str,
        required=False,
        help="Name of site to process (default: all sites are processed)",
    )

    parser.add_argument(
        "-l3", "--folder_l3", type=str, required=True, help="Path to level 3 folder"
    )
    parser.add_argument(
        "-gc",
        "--folder_gcnet",
        type=str,
        required=False,
        help="Path to GC-Net historical L1 folder",
    )

    parser.add_argument(
        "-o",
        "--outpath",
        default=os.getcwd(),
        type=str,
        required=True,
        help="Path where to write output",
    )

    parser.add_argument(
        "-v",
        "--variables",
        default=None,
        type=str,
        required=False,
        help="Path to variables look-up table .csv file for variable name retained" "",
    ),
    parser.add_argument(
        "-m",
        "--metadata",
        default=None,
        type=str,
        required=False,
        help="Path to metadata table .csv file for metadata information" "",
    ),

    args = parser.parse_args(args=debug_args)
    return args


def readNead(infile):
    with open(infile) as f:
        fmt = f.readline()
        assert fmt[0] == "#"
        assert fmt.split("#")[1].split()[0] == "NEAD"
        assert fmt.split("#")[1].split()[1] == "1.0"
        assert fmt.split("#")[1].split()[2] == "UTF-8"

        line = f.readline()
        assert line[0] == "#"
        assert line.split("#")[1].strip() == "[METADATA]"

        meta = {}
        fields = {}
        section = "meta"
        while True:
            line = f.readline()
            if line.strip(" ") == "#":
                continue
            if line == "# [DATA]\n":
                break  # done reading header
            if line == "# [FIELDS]\n":
                section = "fields"
                continue  # done reading header

            if line[0] == "\n":
                continue  # blank line
            assert line[0] == "#"  # if not blank, must start with "#"

            key_eq_val = line.split("#")[1].strip()
            if key_eq_val == "" or key_eq_val == None:
                continue  # Line is just "#" or "# " or "#   #"...
            assert "=" in key_eq_val, print(line, key_eq_val)
            key = key_eq_val.split("=")[0].strip()
            val = key_eq_val.split("=")[1].strip()

            # Convert from string to number if it is a number
            if val.strip("-").strip("+").replace(".", "").isdigit():
                val = float(val)
                if val == int(val):
                    val = int(val)

            if section == "meta":
                meta[key] = val
            if section == "fields":
                fields[key] = val
        # done reading header

        # Find delimiter and fields for reading NEAD as simple CSV
        assert "field_delimiter" in meta.keys()
        assert "fields" in fields.keys()
        FD = meta["field_delimiter"]
        names = [_.strip() for _ in fields.pop("fields").split(FD)]

        df = pd.read_csv(
            infile,
            comment="#",
            names=names,
            sep=FD,
            usecols=np.arange(len(names)),
            skip_blank_lines=True,
        )
        df["timestamp"] = pd.to_datetime(df.timestamp).dt.tz_localize(None)
        df = df.set_index("timestamp")
        ds = df.to_xarray()
        ds.attrs = meta

        # renaming variables
        file_path = pypromice.resources.DEFAULT_VARIABLES_ALIASES_GCNET_PATH
        var_name = pd.read_csv(file_path)
        var_name = var_name.set_index("old_name").GEUS_name
        msk = [v for v in var_name.index if v in ds.data_vars]
        var_name = var_name.loc[msk].to_dict()

        # combining thermocouple and CS100 temperatures
        ds["TA1"] = ds["TA1"].combine_first(ds["TA3"])
        ds["TA2"] = ds["TA2"].combine_first(ds["TA4"])

        # renaming variables to the GEUS names
        ds = ds.rename(var_name)

        # variables always dropped from the historical GC-Net files
        # could be move to the config files at some point
        standard_vars_to_drop = [
            "NR",
            "TA3",
            "TA4",
            "TA5",
            "NR_cor",
            "TA2m",
            "RH2m",
            "VW10m",
            "SZA",
            "SAA",
        ]
        standard_vars_to_drop = standard_vars_to_drop + [
            v for v in list(ds.keys()) if v.endswith("_adj_flag")
        ]

        # Drop the variables if they are present in the dataset
        ds = ds.drop_vars([var for var in standard_vars_to_drop if var in ds])

        ds = ds.rename({"timestamp": "time"})

        # in the historical GC-Net processing, periods with missing z_surf_combined
        # are filled with a constant value, these values should be removed to
        # allow a better alignement with the z_surf_combined estimated for the GEUS stations
        ds["z_surf_combined"] = ds["z_surf_combined"].where(
            ds["z_surf_combined"].diff(dim="time") != 0
        )
    return ds


def loadArr(infile, isNead):
    if infile.split(".")[-1].lower() in "csv":
        if isNead:
            ds = readNead(infile)
        else:
            df = pd.read_csv(infile)
            df["time"] = pd.to_datetime(df["time"]).dt.tz_localize(None)
            df = df.set_index("time")
            ds = xr.Dataset.from_dataframe(df)

    elif infile.split(".")[-1].lower() in "nc":
        with xr.open_dataset(infile) as ds:
            ds.load()
        # Remove encoding attributes from NetCDF
        for varname in ds.variables:
            if ds[varname].encoding != {}:
                ds[varname].encoding = {}

    try:
        name = ds.attrs["station_name"]
    except:
        name = infile.split("/")[-1].split(".")[0].split("_hour")[0].split("_10min")[0]

    print(f"{name} array loaded from {infile}")
    return ds, name


def align_surface_heights(data_series_new, data_series_old):
    """
    Align two surface height time series based on the gap between their end and
    start.

    If the gap between the end of `data_series_old` and the start of `data_series_new`
    is less than a week, the function aligns them based on the median value of
    the last week of `data_series_old` and the first week of `data_series_new`.
    If the gap is larger than a week, it aligns them using a linear fit. If
    there is overlap, the function uses the overlapping period to adjust the
    newer time series.

    Parameters
    ----------
    data_series_old : pandas.Series
        The older time series data.
    data_series_new : pandas.Series
        The newer time series data.

    Returns
    -------
    numpy.ndarray
        Array containing the aligned newer time series data.
    """
    # Get the first and last valid indices of both series
    last_old_idx = data_series_old.last_valid_index()
    first_new_idx = data_series_new.first_valid_index()

    # Check for overlap
    if first_new_idx <= last_old_idx:
        # Find the overlapping period
        overlap_start = first_new_idx
        overlap_end = min(last_old_idx, overlap_start + pd.to_timedelta("7D"))

        # Compute the median values for the overlapping period
        overlap_old = data_series_old[overlap_start:overlap_end].median()
        overlap_new = data_series_new[overlap_start:overlap_end].median()

        if np.isnan(overlap_old) or np.isnan(overlap_new):
            overlap_end = min(last_old_idx, data_series_new.last_valid_index())

            # Compute the median values for the overlapping period
            overlap_old = data_series_old[overlap_start:overlap_end].median()
            overlap_new = data_series_new[overlap_start:overlap_end].median()

        # Align based on the overlapping median values
        data_series_new = data_series_new - overlap_new + overlap_old

    elif (first_new_idx - last_old_idx).days <= 7:
        # Compute the median of the last week of data in the old series
        last_week_old = data_series_old[
            last_old_idx - pd.Timedelta(weeks=1) : last_old_idx
        ].median()

        # Compute the median of the first week of data in the new series
        first_week_new = data_series_new[
            first_new_idx : first_new_idx + pd.Timedelta(weeks=1)
        ].median()

        # Align based on the median values
        data_series_new = data_series_new - first_week_new + last_week_old
    else:
        # Perform a linear fit on the last 5x365x24 non-nan values
        hours_in_5_years = 5 * 365 * 24

        # Drop NaN values and extract the last `hours_in_5_years` non-NaN data points
        data_series_old_nonan = data_series_old.dropna()
        data_series_old_last_5_years = data_series_old_nonan.iloc[
            -min(len(data_series_old), hours_in_5_years):
        ]

        # Perform a linear fit on the last 5 years of data
        fit = np.polyfit(
            data_series_old_last_5_years.index.astype("int64"),
            data_series_old_last_5_years.values,
            1,
        )
        fit_fn = np.poly1d(fit)

        data_series_new = (
            data_series_new.values
            + fit_fn(data_series_new.index.astype("int64")[0])
            - data_series_new[first_new_idx]
        )

    return data_series_new


def build_station_list(config_folder: str, target_station_site: str) -> list:
    """
    Get a list of unique station information dictionaries for a given station site.

    Parameters
    ----------
    config_folder : str
        Path to the folder containing the station configuration TOML files.
    target_station_site : str
        The station site to filter the station information by.

    Returns
    -------
    list
        A list of dictionaries containing station information that have the specified station site.
    """
    station_info_list = []  # Initialize an empty list to store station information

    found_as_station = False
    for filename in os.listdir(config_folder):
        if filename.endswith(".toml"):
            file_path = os.path.join(config_folder, filename)

            with open(file_path, "r") as file:
                data = toml.load(file)  # Load the TOML file
                station_site = data.get("station_site")  # Get the station site
                stid = data.get("stid")  # Get the station ID

                # Check if the station site matches the target and stid is unique
                if stid == target_station_site:
                    found_as_station = True
                if station_site == target_station_site and stid:
                    station_info = data.copy()  # Copy all attributes from the TOML file
                    station_info_list.append(
                        station_info
                    )  # Add the station info to the list

    if len(station_info_list) == 0 and not found_as_station:
        logger.error(
            "\n***\nNo station_configuration file found for %s.\nProcessing it as a single-station PROMICE site.\n***"
            % target_station_site
        )
        station_info = {
            "stid": target_station_site,
            "station_site": target_station_site,
            "project": "PROMICE",
            "location_type": "ice sheet",
        }
        station_info_list.append(station_info)
    elif len(station_info_list) == 0:
        logger.error(
            '\n***\nThe name "%s" passed to join_l3 is a station name and not a site name (e.g. SCO_Lv3 instead of SCO_L). Please provide a site name that is named at least once in the "station_site" attribute of the station configuration files.\n***'
            % target_station_site
        )

    return station_info_list


def join_l3(config_folder, site, folder_l3, folder_gcnet, outpath, variables, metadata):
    # Get the list of station information dictionaries associated with the given site
    list_station_info = build_station_list(config_folder, site)

    # Read the datasets and store them into a list along with their latest timestamp and station info
    list_station_data = []
    for station_info in list_station_info:
        stid = station_info["stid"]

        filepath = os.path.join(folder_l3, stid, stid + "_hour.nc")
        isNead = False
        if station_info["project"].lower() in ["historical gc-net"]:
            filepath = os.path.join(folder_gcnet, stid + ".csv")
            isNead = True
        if not os.path.isfile(filepath):
            logger.error(
                "\n***\n"
                + stid
                + " was listed as station but could not be found in "
                + folder_l3
                + " nor "
                + folder_gcnet
                + "\n***"
            )
            continue

        l3, _ = loadArr(filepath, isNead)

        # removing specific variable from a given file
        specific_vars_to_drop = station_info.get("skipped_variables", [])
        if len(specific_vars_to_drop) > 0:
            logger.info("Skipping %s from %s" % (specific_vars_to_drop, stid))
            l3 = l3.drop_vars([var for var in specific_vars_to_drop if var in l3])

        list_station_data.append((l3, station_info))

    # Sort the list in reverse chronological order so that we start with the latest data
    sorted_list_station_data = sorted(
        list_station_data, key=lambda x: x[0].time.min(), reverse=True
    )
    sorted_stids = [info["stid"] for _, info in sorted_list_station_data]
    logger.info("joining %s" % " ".join(sorted_stids))

    l3_merged = None

    for l3, station_info in sorted_list_station_data:
        stid = station_info["stid"]

        if l3_merged is None:
            # saving attributes of stid
            st_attrs = {}
            st_attrs[stid] = l3.attrs.copy()
            # adding timestamps info
            st_attrs[stid]["first_timestamp"] = (
                l3.time.isel(time=0).dt.strftime(date_format="%Y-%m-%d %H:%M:%S").item()
            )
            st_attrs[stid]["last_timestamp"] = (
                l3.time.isel(time=-1)
                .dt.strftime(date_format="%Y-%m-%d %H:%M:%S")
                .item()
            )

            # then stripping attributes
            attrs_list = list(l3.attrs.keys())
            for k in attrs_list:
                del l3.attrs[k]

            # initializing l3_merged with l3
            l3_merged = l3.copy()

            # creating the station_attributes attribute in l3_merged
            l3_merged.attrs["stations_attributes"] = st_attrs

        else:
            # if l3 (older data) is missing variables compared to l3_merged (newer data)
            # , then we fill them with nan
            for v in l3_merged.data_vars:
                if v not in l3.data_vars:
                    l3[v] = l3.t_u * np.nan
            for v in l3.data_vars:
                if v not in l3_merged.data_vars:
                    l3_merged[v] = l3_merged.t_u * np.nan

            # saving attributes of station under an attribute called $stid
            st_attrs = l3_merged.attrs.get("stations_attributes", {})
            st_attrs[stid] = l3.attrs.copy()
            l3_merged.attrs["stations_attributes"] = st_attrs

            # then stripping attributes
            attrs_list = list(l3.attrs.keys())
            for k in attrs_list:
                del l3.attrs[k]

            l3_merged.attrs["stations_attributes"][stid]["first_timestamp"] = (
                l3.time.isel(time=0).dt.strftime(date_format="%Y-%m-%d %H:%M:%S").item()
            )
            l3_merged.attrs["stations_attributes"][stid]["last_timestamp"] = (
                l3_merged.time.isel(time=0)
                .dt.strftime(date_format="%Y-%m-%d %H:%M:%S")
                .item()
            )

            # adjusting surface height in the most recent data (l3_merged)
            # so that it shows continuity with the older data (l3)
            if "z_surf_combined" in l3_merged.keys() and "z_surf_combined" in l3.keys():
                if (
                    l3_merged.z_surf_combined.notnull().any()
                    and l3.z_surf_combined.notnull().any()
                ):
                    l3_merged["z_surf_combined"] = (
                        "time",
                        align_surface_heights(
                            l3_merged.z_surf_combined.to_series(),
                            l3.z_surf_combined.to_series(),
                        ),
                    )
            if "z_ice_surf" in l3_merged.keys() and "z_ice_surf" in l3.keys():
                if (
                    l3_merged.z_ice_surf.notnull().any()
                    and l3.z_ice_surf.notnull().any()
                ):
                    l3_merged["z_ice_surf"] = (
                        "time",
                        align_surface_heights(
                            l3_merged.z_ice_surf.to_series(), l3.z_ice_surf.to_series()
                        ),
                    )
            
            # saves attributes
            attrs = l3_merged.attrs
            # merging by time block
            l3_merged = xr.concat(
                (
                    l3.sel(
                        time=slice(l3.time.isel(time=0), l3_merged.time.isel(time=0))
                    ),
                    l3_merged,
                ),
                dim="time",
            )
            
            # restauring attributes
            l3_merged.attrs = attrs

    # Assign site id
    if not l3_merged:
        logger.error("No level 3 station data file found for " + site)
        return None, sorted_list_station_data
    l3_merged.attrs["site_id"] = site
    l3_merged.attrs["stations"] = " ".join(sorted_stids)
    l3_merged.attrs["level"] = "L3"
    l3_merged.attrs["project"] = sorted_list_station_data[0][1]["project"]
    l3_merged.attrs["location_type"] = sorted_list_station_data[0][1]["location_type"]

    site_source = dict(
        site_config_source_hash=get_commit_hash_and_check_dirty(config_folder),
        gcnet_source_hash=get_commit_hash_and_check_dirty(folder_gcnet),
    )

    for stid, station_attributes in l3_merged.attrs["stations_attributes"].items():
        if "source" in station_attributes.keys():
            station_source = json.loads(station_attributes["source"])
            for k, v in station_source.items():
                if k in site_source and site_source[k] != v:
                    site_source[k] = "multiple"
                else:
                    site_source[k] = v
    l3_merged.attrs["source"] = json.dumps(site_source)

    v = pypromice.resources.load_variables(variables)
    m = pypromice.resources.load_metadata(metadata)
    if outpath is not None:
        prepare_and_write(l3_merged, outpath, v, m, "60min", nc_compression=True)
        prepare_and_write(l3_merged, outpath, v, m, "1D", nc_compression=True)
        prepare_and_write(l3_merged, outpath, v, m, "M", nc_compression=True)
    return l3_merged, sorted_list_station_data


def main():
    args = parse_arguments_joinl3()
    _, _ = join_l3(
        args.config_folder,
        args.site,
        args.folder_l3,
        args.folder_gcnet,
        args.outpath,
        args.variables,
        args.metadata,
    )


if __name__ == "__main__":
    main()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pypromice",
    version="1.5.0",
    author="GEUS Glaciology and Climate",
    description="PROMICE/GC-Net data processing toolbox",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/GEUS-Glaciology-and-Climate/pypromice",
    project_urls={
        "Bug Tracker": "https://github.com/GEUS-Glaciology-and-Climate/pypromice/issues",
        "Documentation": "https://pypromice.readthedocs.io",
    "Source Code": "https://github.com/GEUS-Glaciology-and-Climate/pypromice"
    },
    keywords="promice gc-net aws climate glaciology greenland geus",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    include_package_data = True,
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.10",
    package_data={
    	"pypromice.tx": ["payload_formats.csv", "payload_types.csv"],
        "pypromice.qc.percentiles": ["thresholds.csv"],
        "pypromice.postprocess": ["positions_seed.csv"],
    },
    install_requires=['numpy~=1.23', 'pandas>=1.5.0', 'xarray>=2022.6.0', 'toml', 'scipy>=1.9.0', 'Bottleneck', 'netcdf4', 'pyDataverse==0.3.1', 'eccodes', 'scikit-learn>=1.1.0'],
#    extras_require={'postprocess': ['eccodes','scikit-learn>=1.1.0']},
    entry_points={
    'console_scripts': [
        'get_promice_data = pypromice.get.get_promice_data:get_promice_data',
        'get_l0tx = pypromice.tx.get_l0tx:get_l0tx',
        'join_l2 = pypromice.process.join_l2:main',
        'join_l3 = pypromice.process.join_l3:main',
        'get_l2 = pypromice.process.get_l2:main',
        'get_l2tol3 = pypromice.process.get_l2tol3:main',
        'make_metadata_csv = pypromice.postprocess.make_metadata_csv:main',
        'get_watsontx = pypromice.tx.get_watsontx:get_watsontx',
        'get_bufr = pypromice.postprocess.get_bufr:main',
        'create_bufr_files = pypromice.postprocess.create_bufr_files:main',
        'bufr_to_csv = pypromice.postprocess.bufr_to_csv:main',
        'get_msg = pypromice.tx.get_msg:get_msg'
    ],
},
)



#!/usr/bin/env python
"""
AWS data processing module
"""
import json
import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)

import logging, os
from pathlib import Path
import pandas as pd
import xarray as xr
from functools import reduce
from importlib import metadata


import pypromice.resources
from pypromice.process.L0toL1 import toL1
from pypromice.process.L1toL2 import toL2
from pypromice.process.L2toL3 import toL3
from pypromice.process import write, load, utilities
from pypromice.utilities.git import get_commit_hash_and_check_dirty

pd.set_option("display.precision", 2)
xr.set_options(keep_attrs=True)
logger = logging.getLogger(__name__)


class AWS(object):
    """AWS object to load and process PROMICE AWS data"""

    def __init__(
        self,
        config_file,
        inpath,
        data_issues_repository: Path | str,
        var_file=None,
        meta_file=None,
    ):
        """Object initialisation

        Parameters
        ----------
        config_file : str
            Configuration file path
        inpath : str
            Input file path
        var_file: str, optional
            Variables look-up table file path. If not given then pypromice's
            variables file is used. The default is None.
        meta_file: str, optional
            Metadata info file path. If not given then pypromice's
            metadata file is used. The default is None.
        """
        assert os.path.isfile(config_file), "cannot find " + config_file
        assert os.path.isdir(inpath), "cannot find " + inpath
        logger.info(
            "AWS("
            f"config_file={config_file},"
            f" inpath={inpath},"
            f" data_issues_repository={data_issues_repository},"
            f" var_file={var_file},"
            f" meta_file={meta_file}"
            ")"
        )

        # Load config, variables CSF standards, and L0 files
        self.config = self.loadConfig(config_file, inpath)
        self.vars = pypromice.resources.load_variables(var_file)
        self.meta = pypromice.resources.load_metadata(meta_file)
        self.data_issues_repository = Path(data_issues_repository)

        config_hash = get_commit_hash_and_check_dirty(Path(config_file))
        config_source_string = f"{Path(config_file).name}:{config_hash}"
        inpath_hash = get_commit_hash_and_check_dirty(Path(inpath))
        data_issues_hash = get_commit_hash_and_check_dirty(self.data_issues_repository)
        source_dict = dict(
            pypromice=metadata.version("pypromice"),
            l0_config_file=config_source_string,
            l0_data_root=inpath_hash,
            data_issues=data_issues_hash,
        )
        logger.debug('Source information: %s', source_dict)
        self.meta["source"] = json.dumps(source_dict)

        # Load config file
        L0 = self.loadL0()
        self.L0 = []
        for l in L0:
            n = write.getColNames(self.vars, l)
            self.L0.append(utilities.popCols(l, n))

        formats = {dataset.attrs["format"].lower() for dataset in self.L0}
        if "raw" in formats:
            self.format = "raw"
        elif "stm" in formats:
            self.format = "STM"
        elif "tx" in formats:
            self.format = "tx"
        else:
            raise ValueError(f"Unknown formats from l0 datasets: {','.join(formats)}")

        self.L1 = None
        self.L1A = None
        self.L2 = None
        self.L3 = None

    def process(self):
        """Perform L0 to L3 data processing"""
        try:
            logger.info(
                f'Commencing {self.L0.attrs["number_of_booms"]}-boom processing...'
            )
            logger.info(
                f'Commencing {self.L0.attrs["number_of_booms"]}-boom processing...'
            )
        except:
            logger.info(
                f'Commencing {self.L0[0].attrs["number_of_booms"]}-boom processing...'
            )
        self.getL1()
        self.getL2()
        self.getL3()

    def getL1(self):
        """Perform L0 to L1 data processing"""
        logger.info("Level 1 processing...")
        self.L0 = [utilities.addBasicMeta(item, self.vars) for item in self.L0]
        self.L1 = [toL1(item, self.vars) for item in self.L0]
        self.L1A = reduce(xr.Dataset.combine_first, reversed(self.L1))
        self.L1A.attrs["format"] = self.format

    def getL2(self):
        """Perform L1 to L2 data processing"""
        logger.info("Level 2 processing...")

        self.L2 = toL2(
            self.L1A,
            vars_df=self.vars,
            data_flags_dir=self.data_issues_repository / "flags",
            data_adjustments_dir=self.data_issues_repository / "adjustments",
        )

    def getL3(self):
        """Perform L2 to L3 data processing, including resampling and metadata
        and attribute population"""
        logger.info("Level 3 processing...")
        self.L3 = toL3(self.L2, data_adjustments_dir=self.data_issues_repository / "adjustments")

    def loadConfig(self, config_file, inpath):
        """Load configuration from .toml file

        Parameters
        ----------
        config_file : str
            TOML file path
        inpath : str
            Input folder directory where L0 files can be found

        Returns
        -------
        conf : dict
            Configuration parameters
        """
        conf = load.getConfig(config_file, inpath)
        return conf

    def loadL0(self):
        """Load level 0 (L0) data from associated TOML-formatted
        config file and L0 data file

        Try readL0file() using the config with msg_lat & msg_lon appended. The
        specific ParserError except will occur when the number of columns in
        the tx file does not match the expected columns. In this case, remove
        msg_lat & msg_lon from the config and call readL0file() again. These
        station files either have no data after Nov 2022 (when msg_lat &
        msg_lon were added to processing), or for whatever reason these fields
        did not exist in the modem message and were not added.

        Returns
        -------
        ds_list : list
            List of L0 xr.Dataset objects
        """
        ds_list = []
        for k in self.config.keys():
            target = self.config[k]
            try:
                ds_list.append(self.readL0file(target))

            except pd.errors.ParserError as e:
                # ParserError: Too many columns specified: expected 40 and found 38
                # logger.info(f'-----> No msg_lat or msg_lon for {k}')
                for item in ["msg_lat", "msg_lon"]:
                    target["columns"].remove(item)  # Also removes from self.config
                ds_list.append(self.readL0file(target))
            logger.info(f"L0 data successfully loaded from {k}")
        return ds_list

    def readL0file(self, conf):
        """Read L0 .txt file to Dataset object using config dictionary and
        populate with initial metadata

        Parameters
        ----------
        conf : dict
            Configuration parameters

        Returns
        -------
        ds : xr.Dataset
            L0 data
        """
        file_version = conf.get("file_version", -1)
        ds = load.getL0(
            conf["file"],
            conf["nodata"],
            conf["columns"],
            conf["skiprows"],
            file_version,
            time_offset=conf.get("time_offset"),
        )
        ds = utilities.populateMeta(ds, conf, ["columns", "skiprows", "modem"])
        return ds
